#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the garoupa project.
#  Please respect the license - more about this in the section (*) below.
#
#  garoupa is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  garoupa is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with garoupa.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.


def extend_history(d, hosh):
    """
    >>> from ldict import ldict
    >>> d = ldict()
    >>> d.history
    {}
    >>> d["x"] = 5
    >>> from ldict import decolorize
    >>> decolorize(str(d.history)) == '{Tz_d158c49297834fad67e6de7cdba3ea368aae4: None}'
    True
    >>> d["y"] = 7
    >>> decolorize(str(d.history)) == '{I0_39c94b4dfbc7a8579ca1304eba25917204a5e: {Tz_d158c49297834fad67e6de7cdba3ea368aae4, Rs_92162dea64a7462725cac7dcee71b67669f69}}'
    True

    Parameters
    ----------
    d
    hosh

    Returns
    -------

    """
    if not d.history or hosh.etype == "ordered" or d.last and "_" not in d.last.id[:3]:
        d.last = hosh
        d.history[d.last] = None
    else:
        many = d.history.pop(d.last)
        previous_last = d.last
        d.last *= hosh
        if many is None:
            d.history[d.last] = {previous_last, hosh}
        else:
            many.add(hosh)
            d.history[d.last] = many


def rewrite_history(d, hosh):
    """
    >>> from ldict import ldict
    >>> d = ldict()
    >>> d["x"] = 5
    >>> del d["x"]
    >>> list(d.ids.keys()) == []
    True
    >>> d["x"] = 5
    >>> d["y"] = 7
    >>> from ldict import decolorize
    >>> decolorize(str(d.history)) == '{I0_39c94b4dfbc7a8579ca1304eba25917204a5e: {Tz_d158c49297834fad67e6de7cdba3ea368aae4, Rs_92162dea64a7462725cac7dcee71b67669f69}}'
    True
    >>> d["z"] = 8
    >>> decolorize(str(d.history)) == '{-H_b4adb0fb38e343cfe5984eec6d1c92907539b: {hH_2713d4ae3c47207859e6cd9ea2f6013e6193d, Tz_d158c49297834fad67e6de7cdba3ea368aae4, Rs_92162dea64a7462725cac7dcee71b67669f69}}'
    True
    >>> list(d.ids.keys()) == ["x", "y", "z"]
    True
    >>> del d["x"]
    >>> list(d.ids.keys()) == ["y", "z"]
    True
    >>> del d["z"]
    >>> decolorize(str(d.history)) == '{Rs_92162dea64a7462725cac7dcee71b67669f69: None}'
    True
    >>> d >>= lambda y: {"x": y**2}
    >>> d >>= lambda y: {"w": y**2}
    >>> del d["x"]
    >>> del d["w"]
    >>> decolorize(str(d.history)) == '{Rs_92162dea64a7462725cac7dcee71b67669f69: None}'
    True
    >>> d["u"] = 2
    >>> d >>= lambda y: {"w": y**2}
    >>> d["v"] = 8
    >>> decolorize(str(d.history)) == '{j9_045f5ac950e6b2ff74f26c099994c606a119b: {Rs_92162dea64a7462725cac7dcee71b67669f69, uI_acddd31feb9259465f280b47ba1310d928122}, MOQ356EM3Aw3fpX8bh3lbZBjdhuGHJTLdHHU130o: None, dH_b1735f953d1c9a6859e61eaea2f6012e6193d: None}'
    True
    >>> d["d"] = d
    >>> d.show(colored=False)
    {
        "id": "il-YWAqTLqxXMPK7BGjEe1rQra32prLvrinN360M",
        "ids": {
            "w": "yFRT.QXWbEBxKPv6.HY0WTtTUxrGHJTLdHHU130o",
            "y": "Rs_92162dea64a7462725cac7dcee71b67669f69",
            "u": "uI_acddd31feb9259465f280b47ba1310d928122",
            "v": "dH_b1735f953d1c9a6859e61eaea2f6012e6193d",
            "d": "9SjmNApbfmSKqS051DRVOda9hlGGHJTLdHHU130o"
        },
        "w": "→(y)",
        "y": 7,
        "u": 2,
        "v": 8,
        "d": {
            "id": "BlZvumbNqO9tdyGB3f8wXY0GhGEGHJTLdHHU130o",
            "ids": {
                "w": "yFRT.QXWbEBxKPv6.HY0WTtTUxrGHJTLdHHU130o",
                "y": "Rs_92162dea64a7462725cac7dcee71b67669f69",
                "u": "uI_acddd31feb9259465f280b47ba1310d928122",
                "v": "dH_b1735f953d1c9a6859e61eaea2f6012e6193d"
            },
            "w": "→(y)",
            "y": 7,
            "u": 2,
            "v": 8
        }
    }
    >>> d["s"] = 1
    >>> del d["u"]
    >>> d["p"] = 82
    >>> d["q"] = 18
    >>> from ldict import decolorize
    >>> decolorize(str(d.history)) == '{yFRT.QXWbEBxKPv6.HY0WTtTUxrGHJTLdHHU130o: None, 38_19e4d550a16004128ea17f609178b75aea8a6: {dH_b1735f953d1c9a6859e61eaea2f6012e6193d, Rs_92162dea64a7462725cac7dcee71b67669f69}, 9SjmNApbfmSKqS051DRVOda9hlGGHJTLdHHU130o: None, wo_dbc5bf9d6d53cdc50c64e67323f8a3f8bb04d: {pt_369001ff096b860c63702a81e7bb7dc6252ae, Sr_028bb0aa9a6ddd4c6527ed7a94fd3322b9bfb, gv_06cc5343c944510b23dca462b64fe26acc2a2}}'
    True

    Parameters
    ----------
    d
    key

    Returns
    -------

    """
    lastitem = d.last and d.history[d.last]
    if d.last == hosh:
        del d.history[d.last]
        hist = list(d.history.keys())
        d.last = hist[-1] if len(hist) > 0 else None
    elif isinstance(lastitem, set) and hosh in lastitem:  # TODO: check inside list?
        lastitem.remove(hosh)
        del d.history[d.last]
        d.last /= hosh
        d.history[d.last] = lastitem if len(lastitem) > 1 else None  # pragma: no cover
    else:
        d.history = {}
        hosh = d.identity.h
        many = set()
        for k, v in d.hoshes.items():
            if v.etype == "ordered":
                if hosh != d.identity.h:
                    d.history[hosh] = None if len(many) == 1 else many
                    hosh = d.identity.h
                    many = set()
                d.history[v] = None
            else:
                hosh *= v
                many.add(v)
        if hosh != d.identity.h:
            d.history[hosh] = None if len(many) == 1 else many
        d.last = list(d.history.keys())[-1]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ldict']

package_data = \
{'': ['*']}

install_requires = \
['garoupa>=2.210907.7,<3.0.0',
 'orjson>=3.5.0,<4.0.0',
 'pdoc3>=0.10.0,<0.11.0',
 'uncompyle6>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'ldict',
    'version': '2.210908.4',
    'description': 'Lazy dict with universally unique identifier for values',
    'long_description': '![test](https://github.com/davips/ldict/workflows/test/badge.svg)\n[![codecov](https://codecov.io/gh/davips/ldict/branch/main/graph/badge.svg)](https://codecov.io/gh/davips/ldict)\n\n# ldict\nUniquely identified lazy dict.\n\n[Latest version as a package](https://pypi.org/project/ldict)\n\n[Current code](https://github.com/davips/ldict)\n\n## Overview\nWe consider that every data is generated by a process, starting from `empty`.\nThe process is a sequence of transformation steps that can be of two types:\nvalue insertion and function application.\nValue insertion is done using dict-like objects as shown below.\nThe operator `>>` concatenate the steps chronologically.\n![img.png](https://raw.githubusercontent.com/davips/ldict/main/examples/img.png)\n\nFunction application is done in the same way.\nThe parameter names define the input fields,\nwhile the keys in the returned dict define the output fields:\n![img_1.png](https://raw.githubusercontent.com/davips/ldict/main/examples/img_1.png)\n\nSimilarly, for anonymous functions:\n![img_5.png](https://raw.githubusercontent.com/davips/ldict/main/examples/img_5.png)\n\nFinally, the result is only evaluated at request:\n![img_6.png](https://raw.githubusercontent.com/davips/ldict/main/examples/img_6.png)\n![img_7.png](https://raw.githubusercontent.com/davips/ldict/main/examples/img_7.png)\n\n\n## Installation\n### ...as a standalone lib\n```bash\n# Set up a virtualenv. \npython3 -m venv venv\nsource venv/bin/activate\n\n# Install from PyPI...\npip install --upgrade pip\npip install -U ldict\n\n# ...or, install from updated source code.\npip install git+https://github.com/davips/ldict\n```\n\n### ...from source\n```bash\ngit clone https://github.com/davips/ldict\ncd ldict\npoetry install\n```\n\n## Examples\n**Merging two ldicts**\n<details>\n<p>\n\n```python3\nfrom ldict import ldict\n\na = ldict(x=3)\nprint(a)\n"""\n{\n    "id": "kr_4aee5c3bcac2c478be9901d57fd1ef8a9d002",\n    "ids": "kr_4aee5c3bcac2c478be9901d57fd1ef8a9d002",\n    "x": 3\n}\n"""\n```\n\n```python3\n\nb = ldict(y=5)\nprint(b)\n"""\n{\n    "id": "Uz_0af6d78f77734fad67e6de7cdba3ea368aae4",\n    "ids": "Uz_0af6d78f77734fad67e6de7cdba3ea368aae4",\n    "y": 5\n}\n"""\n```\n\n```python3\n\nprint(a >> b)\n"""\n{\n    "id": "c._2b0434ca422114262680df425b85cac028be6",\n    "ids": "kr_4aee5c3bcac2c478be9901d57fd1ef8a9d002 Uz_0af6d78f77734fad67e6de7cdba3ea368aae4",\n    "x": 3,\n    "y": 5\n}\n"""\n```\n\n\n</p>\n</details>\n\n**Lazily applying functions to ldict**\n<details>\n<p>\n\n```python3\nfrom ldict import ldict\n\na = ldict(x=3)\nprint(a)\n"""\n{\n    "id": "kr_4aee5c3bcac2c478be9901d57fd1ef8a9d002",\n    "ids": "kr_4aee5c3bcac2c478be9901d57fd1ef8a9d002",\n    "x": 3\n}\n"""\n```\n\n```python3\n\na = a >> ldict(y=5) >> {"z": 7} >> (lambda x, y, z: {"r": x ** y // z})\nprint(a)\n"""\n{\n    "id": "8jopGVdtSEyCk1NSKcrEF-Lfv8up9MQBdvkLxU2o",\n    "ids": "J3tsy4vUXPELySBicaAy-h-UK7Dp9MQBdvkLxU2o... +2 ...Ss_7dff0a161ba7462725cac7dcee71b67669f69",\n    "r": "→(x y z)",\n    "x": 3,\n    "y": 5,\n    "z": 7\n}\n"""\n```\n\n```python3\n\nprint(a.r)\n"""\n34\n"""\n```\n\n```python3\n\nprint(a)\n"""\n{\n    "id": "8jopGVdtSEyCk1NSKcrEF-Lfv8up9MQBdvkLxU2o",\n    "ids": "J3tsy4vUXPELySBicaAy-h-UK7Dp9MQBdvkLxU2o... +2 ...Ss_7dff0a161ba7462725cac7dcee71b67669f69",\n    "r": 34,\n    "x": 3,\n    "y": 5,\n    "z": 7\n}\n"""\n```\n\n\n</p>\n</details>\n\n## Persistence\nExtra dependencies can be installed to support saving data to disk or to a server in the network. \n\n**[still an ongoing work...]**\n\n`poetry install -E full`\n\n## Concept\nA ldict is like a common Python dict, with extra funtionality and lazy.\nIt is a mapping between string keys, called fields, and any serializable object.\nThe ldict `id` (identifier) and the field `ids` are also part of the mapping.  \n\nThe user can provide a unique identifier ([hosh](https://pypi.org/project/garoupa))\nfor each function or value object.\nOtherwise, they will be calculated through blake3 hashing of the content of data or bytecode of function.\nFor this reason, such functions should be simple, i.e.,\nwith minimal external dependencies, to avoid the unfortunate situation where two\nfunctions with identical local code actually perform different calculations through\ncalls to external code that implement different algorithms with the same name.\nAlternatively, a Hosh object can be passed inside the dict that is returned by the function, under the key "_id".\n\n## Usage\n\n**[still an ongoing work...]**\n',
    'author': 'davips',
    'author_email': 'dpsabc@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .appearance import decolorize
from .ldict_ import Ldict

ldict = Ldict

empty = ldict(readonly=True)
"""The empty object is used to induce a ldict from a dict"""

ø = empty
"""UTF-8 alias for the empty object, it is used to induce a ldict from a dict. AltGr+O in most keyboards."""

decolorize = decolorize
